package zero1hd.rhythmbullet.stages;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Preferences;
import com.badlogic.gdx.assets.AssetManager;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.Pixmap.Format;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.graphics.glutils.FrameBuffer;
import com.badlogic.gdx.graphics.glutils.ShaderProgram;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.utils.viewport.FitViewport;

import zero1hd.rhythmbullet.RhythmBullet;
import zero1hd.rhythmbullet.audio.map.EntitySpawnInfo;
import zero1hd.rhythmbullet.audio.map.GamePlayMap;
import zero1hd.rhythmbullet.audio.map.MapWindowData;
import zero1hd.rhythmbullet.controls.KeyMap;
import zero1hd.rhythmbullet.entity.CollisionDetector;
import zero1hd.rhythmbullet.entity.Entity;
import zero1hd.rhythmbullet.entity.EntityManager;
import zero1hd.rhythmbullet.entity.ally.Laser;
import zero1hd.rhythmbullet.entity.ally.PolyjetEntity;
import zero1hd.rhythmbullet.entity.coordinator.CoordinatorManager;
import zero1hd.rhythmbullet.util.ScoreManager;


public class GamePlayArea extends Stage {
	public PolyjetEntity polyjet;
	private GamePlayMap audioMap;
	
	public CoordinatorManager cm;
	public EntityManager em;
	private CollisionDetector collisionDetector;
	
	public ScoreManager score = new ScoreManager();
	
	private ShaderProgram gaussianBlurShader;
	private ShaderProgram brightFilterShader;
	private ShaderProgram combineShader;
	private FrameBuffer lightFilterBuffer;
	private FrameBuffer normalBuffer;
	private FrameBuffer hBlur, vBlur;
	private TextureRegion fboRegion;
	private int fboSize;
	
	public GamePlayArea(AssetManager assetManager, Preferences prefs) {
		super(new FitViewport(RhythmBullet.GAME_AREA_WIDTH, RhythmBullet.GAME_AREA_HEIGHT));
		Gdx.app.debug("Game Area", "new area created");
		
		polyjet = new PolyjetEntity(assetManager, 25f, 25f, 100, "standard");
		em = new EntityManager(assetManager, prefs, this);
		cm = new CoordinatorManager(em);
		
		collisionDetector = new CollisionDetector(em.activeEnemies, em.activeAllies, assetManager, prefs);
		em.activeAllies.add(polyjet);
		addActor(polyjet);
	}
	
	public void setAudioMap(GamePlayMap audioMap) {
		this.audioMap = audioMap;
	}
	
	public GamePlayMap getAudioMap() {
		return audioMap;
	}
	
	/**
	 * needs to be called right after set as screen (should be called in show method).
	 * This is due to the saving preference being done once the screen is hidden.
	 * @param prefs
	 */
	public void loadShaders(Preferences prefs) {
		if (prefs.getBoolean("glow shader")) {
			Gdx.app.debug("Shader", "using glow shader");
			brightFilterShader = new ShaderProgram(Gdx.files.internal("shaders/basic.vsh"), Gdx.files.internal("shaders/bright_filter.fsh"));
			if (!brightFilterShader.isCompiled()) {
				System.err.println(brightFilterShader.getLog());
				System.exit(0);
			}
			if (brightFilterShader.getLog().length() != 0) {
				System.out.println(brightFilterShader.getLog());
			}
				
			gaussianBlurShader = new ShaderProgram(Gdx.files.internal("shaders/basic.vsh"), Gdx.files.internal("shaders/gaussian_blur.fsh"));
			if (!gaussianBlurShader.isCompiled()) {
				System.err.println(gaussianBlurShader.getLog());
				System.exit(0);
			}
			if (gaussianBlurShader.getLog().length() != 0) {
				System.out.println(gaussianBlurShader.getLog());
			}
			
			combineShader = new ShaderProgram(Gdx.files.internal("shaders/basic.vsh"), Gdx.files.internal("shaders/combine.fsh"));
			if (!combineShader.isCompiled()) {
				System.err.println(combineShader.getLog());
				System.exit(0);
			}
			if (combineShader.getLog().length() != 0) {
				System.out.println(combineShader.getLog());
			}
			
			
			if (Gdx.graphics.getWidth() < 1024) {
				fboSize = 2048;
			} else if (Gdx.graphics.getWidth() < 2048) {
				fboSize = 4096;
			} else {
				fboSize = 4096;
			}
			
			lightFilterBuffer = new FrameBuffer(Format.RGBA8888, fboSize/4, fboSize/4, false);
			normalBuffer = new FrameBuffer(Format.RGBA8888, fboSize, fboSize, false);
			hBlur = new FrameBuffer(Format.RGBA8888, fboSize, fboSize, false);
			vBlur = new FrameBuffer(Format.RGBA8888, fboSize, fboSize, false);
			
			fboRegion = new TextureRegion();
			
			combineShader.begin();
			combineShader.setUniformi("u_texture1", 1);
			combineShader.end();
			
			vBlur.getColorBufferTexture().bind(1);
			
			Gdx.gl.glActiveTexture(GL20.GL_TEXTURE0);
		}
		
		ShaderProgram.pedantic = false;
	}
	
	@Override
	public void draw() {
		if (gaussianBlurShader != null) {
			//Begin drawing a normal version of screen
			normalBuffer.begin();
			Gdx.gl.glClearColor(0f, 0f, 0f, 0f);
			Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);
			super.draw();
			collisionDetector.renderParticles(getBatch(), Gdx.graphics.getDeltaTime(), true);
			normalBuffer.end(getViewport().getScreenX(), getViewport().getScreenY(), getViewport().getScreenWidth(), getViewport().getScreenHeight());
			//END
			
			//Begin light filtering
			lightFilterBuffer.begin();
			Gdx.gl.glClearColor(0f, 0f, 0f, 0f);
			Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);
			getBatch().setShader(brightFilterShader);
			getBatch().begin(); //BATCH STARTS HERE
			getRoot().draw(getBatch(), 1f);
			getBatch().flush();
			lightFilterBuffer.end(getViewport().getScreenX(), getViewport().getScreenY(), getViewport().getScreenWidth(), getViewport().getScreenHeight());
			
			//Horizontal gaussian blur
			hBlur.begin();
			fboRegion.setTexture(lightFilterBuffer.getColorBufferTexture());
			getBatch().setShader(gaussianBlurShader);
			gaussianBlurShader.setUniformf("horizontal", 1);
			getBatch().draw(fboRegion, 0f, 0f, RhythmBullet.GAME_AREA_WIDTH, RhythmBullet.GAME_AREA_HEIGHT);
			getBatch().flush();
			hBlur.end(getViewport().getScreenX(), getViewport().getScreenY(), getViewport().getScreenWidth(), getViewport().getScreenHeight());
			
			//Vertical gaussian blur
			vBlur.begin();
			fboRegion.setTexture(hBlur.getColorBufferTexture());
			getBatch().setShader(gaussianBlurShader);
			gaussianBlurShader.setUniformf("horizontal", 0);
			getBatch().draw(fboRegion, 0f, 0f, RhythmBullet.GAME_AREA_WIDTH, RhythmBullet.GAME_AREA_HEIGHT);
			getBatch().flush();
			vBlur.end(getViewport().getScreenX(), getViewport().getScreenY(), getViewport().getScreenWidth(), getViewport().getScreenHeight());
			
			//Draw everything to screen
			getBatch().setShader(combineShader);
			fboRegion.setRegion(normalBuffer.getColorBufferTexture());
			fboRegion.flip(false, true);
			getBatch().draw(fboRegion, 0f, 0f, RhythmBullet.GAME_AREA_WIDTH, RhythmBullet.GAME_AREA_HEIGHT);
			getBatch().setShader(null);
			getBatch().end(); //BATCH ENDS HERE
			
		} else {
			super.draw();
		}
	}
	
	@Override
	public void act(float delta) {
		MapWindowData mwd;
		if (audioMap != null && audioMap.getMusicData().getPlaybackMusic().isPlaying()) {
			audioMap.getMusicData().readIndexUpdate();
			if ((mwd = audioMap.getCurrentWindowBasedOnIndex()) != null) {
				EntitySpawnInfo[] currentSpawnInfo = mwd.getArray();
				if (currentSpawnInfo != null) {
					for (int i = 0; i < currentSpawnInfo.length; i++) {
						Entity entity = currentSpawnInfo[i].getEntityToSpawn().buildEntity();
						if (currentSpawnInfo[i].getEntityCoordinator() != null) {
							entity.setCoordinator(currentSpawnInfo[i].getEntityCoordinator().buildCoordinator());
						}

						entity.init(currentSpawnInfo[i].parameters);
						addActor(entity);
					}
				}
			}
		}

		collisionDetector.collisionCheck();
		score.addScore(collisionDetector.getAmassedPoints());
		
		super.act(delta);
	}
	
	@Override
	public boolean keyDown(int keycode) {
		if (keycode == KeyMap.left) {
			polyjet.moveLeft = true;
		}
		if (keycode == KeyMap.right) {
			polyjet.moveRight = true;
		}
		
		if (keycode == KeyMap.up) {
			polyjet.moveUp = true;
		}
		
		if (keycode == KeyMap.down) {        
			polyjet.moveDown = true;
		}
		
		if (keycode == KeyMap.accelerate) {
			polyjet.accelerate = true;
		}
		return false;
	}
	
	@Override
	public boolean keyUp(int keycode) {
		if (keycode == KeyMap.left) {
			polyjet.moveLeft = false;
		}
		if (keycode == KeyMap.right) {
			polyjet.moveRight = false;
		}
		if (keycode == KeyMap.up) {
			polyjet.moveUp = false;
		}
		if (keycode == KeyMap.down) {
			polyjet.moveDown = false;
		}
		
		if (keycode == KeyMap.accelerate) {
			polyjet.accelerate = false;
		}
		
		if (keycode == KeyMap.shoot) {
			Laser laser	= em.laser.buildEntity();
			laser.init(polyjet.getX() + polyjet.getWidth()/2f, polyjet.getY() + polyjet.getHeight()+1f, 60f);
			addActor(laser);
		}
		return false;
	}

	public PolyjetEntity getPolyjetEntity() {
		return polyjet;
	}
	
	@Override
	public void dispose() {
		if (gaussianBlurShader != null) {
			normalBuffer.dispose();
			gaussianBlurShader.dispose();
		}
		super.dispose();
	}
}
